#!/usr/bin/python

"""
analyze_3lambda.py: Analyzes itu_channel/wavelength/crtt that was recorded and
stored in a file using wr_serial_3lamda.py

-------------------------------------------------------------------------------
Copyright (C) 2017 Peter Jansweijer, Henk Peek

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>
-------------------------------------------------------------------------------

Usage:
  analyze_3lambda.py     <root-name> (for example "180928_12_09_18_3_lambda_insitu_alpha_scan")
  analyze_3lambda.py     -h | --help

  <root-name> identifies the files to analyse. Four files have the root-name plus:
              <root-name>_l1_a    lambda 1 (-> lambda 2)
              <root-name>_l2_a    lambda 2
              <root-name>_l2_b    lambda 2 (-> lambda 1)
              <root-name>_l1_b    lambda 1
              so two sets of files are recorded: 'a' setting first lambda 1 then lambda 2
              and 'b' setting first lambda 2 then lambda 1. This is to evaluate whether
              tuning in one or the other direction is of influence on the CRTT measurements.
              So far this has not been an issue.

Options:
  -h --help    Show this screen.
  -t           tolerance [ps], skip points outside fitted line, default: "200"
  -i           use ITU channels instead of wavelength

"""

import os
import sys
import scipy
from scipy.optimize import curve_fit
import numpy
import matplotlib.pyplot as plt

import pdb

# Add parent directory (containing 'lib') to the module search path
lib_path = (os.path.dirname(os.path.abspath(__file__)))
lib_path = os.path.join(lib_path,"..")
sys.path.insert(0,lib_path)

# private imports:
import lib.itu_conversions as itu_conv
import lib.wrt_sfppt015sc as tunable

############################################################################



def spec_alpha(alpha):
  """
  returns the value of alpha as it is used by the SPEC
  (see also White Rabbit Calibration procudeure:
  http://www.ohwr.org/attachments/4092/WR_Calibration-v1.1-20151109.pdf
  equation 11.

  alpha         -- <float> alpha value

  returns:
  alpha         -- <float> alpha according to SPEC 

  """
  return(2**40 * (((alpha+1)/(alpha+2))-0.5))

############################################################################

def calc_alpha_l1(lambda_1, lambda_2, crtt_lambda_1, crtt_lambda_2, fixed_lambda):
  """
  calc_alpha_l1 calculates the alpha factor for wavelength lambda l1.
  For this calculation we need two cable round trip times that were measured
  using lamdba l1 and lambda l2 while both used a common fixed_lambda for the
  return channel.

  lambda_1, lambda_2, fixed_lambda  -- <float> lambda's in [m]
  crtt_l1, crtt_l2                  -- <float> cable round trip times [ps] measured
                                               at lambda_1 and lambda_2

  returns:
  alpha_lambda_1                    -- <float>

  """

  delta_lambda_1 = lambda_1 - fixed_lambda
  delta_lambda_2 = lambda_2 - fixed_lambda
  crtt_diff      = crtt_lambda_1 - crtt_lambda_2
  lambda_diff    = lambda_1 - lambda_2

  alpha_lambda_1 = (2 * delta_lambda_1 * crtt_diff)/((crtt_lambda_1 * lambda_diff) - (crtt_diff * delta_lambda_1))
  return (alpha_lambda_1)

############################################################################

def calc_alpha_tangent(l, crtt_l, tangent, fixed_lambda):
  """
  calc_alpha_tangent calculates the alpha factor for the wavelength lambda l.
  It takes the cable rount trip time that was measured at this lambda
  (crtt_l) while using fixed_lambda for the return path.
  The tangent is the slope of variation in time over wavelength.

  l                 -- <float> lambda [m]
  crtt_l            -- <int>   cable round trip time measured at lambda l
                               using fixed_lambda on the return channel [ps]
  tangent           -- <float> [ps/nm]
  fixed_lambda      -- <float> fixed return channel lambda [m]

  returns:
  alpha             -- <float> alpha value for lambda l when using fixed_lambda
                               as return channel

  """

  delta_l = (l - fixed_lambda)
  alpha = (2* tangent * delta_l )/(crtt_l - tangent * delta_l )

  return(alpha)

############################################################################

def calc_alpha_sellmeier(l, fixed_lambda, A, B, C, D, E):
  """
  calc_alpha_sellmeier calculates the alpha factor for the wavelength lambda l
  while using fixed_lambda for the return path.
  It takes the Sellmeier cooefcients that came out of a cable rount trip time
  wavelength scan.

  l                 -- <float> lambda [m]
  fixed_lambda      -- <float> fixed return channel lambda [m]
  a..e              -- <float> Sellmeier terms

  returns:
  alpha             -- <float> alpha value for lambda l when using fixed_lambda
                               as return channel

  """

  bb = l**2 - fixed_lambda**2
  cc = l**-2 - fixed_lambda**-2
  dd = l**4 - fixed_lambda**4
  ee = l**-4 - fixed_lambda**-4

  alpha = 2* (B*bb + C*cc + D*dd + E*ee)/(A + B*fixed_lambda**2 + C*fixed_lambda**-2 + D*fixed_lambda**4 + E*fixed_lambda**-4)

  return(alpha)


############################################################################

def calc_alpha_error(alpha1, alpha2, crtt_fixed_lambda):
  """
  calc_alpha_error calculates the timing error between the PPS-ses of Master
  and Slave that will be the result of an "imperfect" alpha2 with respect to a
  "perfect" alpha1.
  The timing error scales linear with the round trip time (i.e. the length of
  your fiber)

  alpha1            -- <float> alpha to be considered the correct value
  alpha2            -- <float> alpha for which timing error needs to be calculated
  crtt_fixed_lambda -- <float> crtt value at the fixed wavelenght.
                               note that this is the virtual crtt for which
                               master2slave and slave2master wavelengths would be
                               equal. It can be estimated by extrapolating the crtt
                               array.

  returns:
  t_err             -- <float> PPS time error

  """

  t_err = ((alpha1-alpha2)*crtt_fixed_lambda)/4

  return(t_err)

############################################################################


def line(x, a, b):
  return a * x + b

############################################################################

def group_delay(l, A, B, C, D, E):
  """
  group_delay calculates the group delay for the wavelength lambda l based on
  the 5-term Sellmeier equation with coefficients A..E.
  
  5-term sellmeier group delay:
  tau(lambda) = A + B * lambda^2 + C * lambda^-2 + D * lambda^4 + E * lambda^-4

  A..E              -- <numpy.float64> Sellmeier coefficients
  l                 -- <float> lambda is in [nm].
                    -- Note! Convert to [m] before executing sellmeier fit

  returns:
  delay             -- <float> group delay value based on selm5 and l
  """

  # l = lambda is in [nm]. convert to [m] before executing sellmeier fit
  l = l/1e9
  sellmeier_5 = A + B * l**2 + C * l**-2 + D * l**4 + E * l**-4
  # and convert fitted value back to [nm]
  return sellmeier_5 * 1e9

############################################################################


def read_file(filename):
  """
  read_file
  
  filename          -- pointer to an open datafile

  returns:
  data_dict         -- <dict> with key "ITU-Lambda1 = ##.#, ITU-Lambda2 = ##.#"
                              and values:
                                number of measurements
                                mean crtt value
                                stddev of crtt values
  """

  data_file = open(os.path.join(filename),"r")
  line = data_file.readline()
  if line[:len("#3-lambda measurement for")] != "#3-lambda measurement for":
    print(filename + " is not a 3-lambda measurement file.")
    data_file.close()
    sys.exit()

  line = data_file.readline()	# Read date
  line = data_file.readline()	# Read time
  line = data_file.readline()	# Read comment

  crtt = []
  ITU_l1 = ""
  ITU_l2 = ""
  key = ""
  # create an empty dictionary for storing mean and stddev crtt's
  data_dict = {}
  
  while 1:
    line = data_file.readline()
    if line == "":
      break #end of file
    
    # files look like this:
    #sfp_channel, ITU channel, ITU wavelength, crtt [ps], drxm, drxs, fiber-spool-temp, wr-slave-temp, meas-time
    #ITU-Lambda1 = 18.0, ITU-Lambda2 = 19.0
    #15, 18.0, 1.5630472262773723e-06, 492049342.6666667, 4000, 800, 24.499611, 47.4375, 15:49:26
    if line[:len("ITU-Lambda1")] == "ITU-Lambda1":  # Get Lambda 1 and 2 ITU channel numbers
      line_lst = line.strip().split(",")
      if len(crtt) != 0:
        crtt_arr=numpy.array(crtt)
        crtt_mean = crtt_arr.mean(axis=0)
        crtt_std  = crtt_arr.std(axis=0,ddof=1)
        #print ("ITU: ", ITU_l1, ITU_l2)
        #print("crtt:", len(crtt), crtt_mean, crtt_std)
        data_dict[key]= len(crtt), crtt_mean, crtt_std
        crtt = [] # clean and ready for the next set
      key = line.strip()
      ITU_l1 = float(line_lst[0][len("ITU Lambda1 = "):])
      ITU_l2 = float(line_lst[1][len("ITU Lambda2 = "):])
    else:
      line_lst = line.strip().split(",")
      crtt.append(float(line_lst[3])) # 3-rd item is CRTT

  # catch the last accumulated crtt's
  if len(crtt) != 0:
    crtt_arr=numpy.array(crtt)
    crtt_mean = crtt_arr.mean(axis=0)
    crtt_std  = crtt_arr.std(axis=0,ddof=1)
    #print ("ITU: ", ITU_l1, ITU_l2)
    #print("crtt:", len(crtt), crtt_mean, crtt_std)
    data_dict[key]= len(crtt), crtt_mean, crtt_std

  return (data_dict)

###########################################################################
#
# If run from commandline, we can test the library
#
"""
Usage:
  analyze_3lambda.py     <root-name> (for example "180928_12_09_18_3_lambda_insitu_alpha_scan")
  analyze_3lambda.py     -h | --help

  <root-name> identifies the files to analyse. Four files have the root-name plus:
              <root-name>_l1_a    lambda 1 (-> lambda 2)
              <root-name>_l2_a    lambda 2
              <root-name>_l2_b    lambda 2 (-> lambda 1)
              <root-name>_l1_b    lambda 1
              so two sets of files are recorded: 'a' setting first lambda 1 then lambda 2
              and 'b' setting first lambda 2 then lambda 1. This is to evaluate whether
              tuning in one or the other direction is of influence on the CRTT measurements.
              So far this has not been an issue.

Options:
  -h --help    Show this screen.
  -t           tolerance [ps], skip points outside fitted line, default: "200"
  -i           use ITU channels instead of wavelength

"""
if __name__ == "__main__":

  import argparse
  parser = argparse.ArgumentParser()
  parser.add_argument("name", help="root-name for _l1_a, _l2_a, _l1_b, _l2_b files")
  args = parser.parse_args()

  name = args.name
  fixed_itu_channel = 20
  fixed_lambda = itu_conv.itu_2_wavelength(fixed_itu_channel)

  files = []
  # Keep tangent and temperature values in an array in case multiple
  # files are processed.
  tangent_array = numpy.array([])
  temperature_array = numpy.array([])

  # Figure out if the input files exist. Extensions should be:
  files_exist = True
  ext = ["_l1_a", "_l2_a", "_l1_b", "_l2_b"]
  for extension in ext:
    if os.path.isfile(name + extension) == False:
      files_exist = False

  if files_exist:
    print("files found:")
  else:
    print(name+": does not exist")
    sys.exit()

  # analyse the set of four files files
  # create an empty dictionary for storing mean and stddev crtt's for each of the four files
  data_set={}
  for extension in sorted(ext):
    try:
      print (name + extension)
      data_set[extension]=read_file(name + extension)
    except:
      print (name + extension + ": wrong file type")
      continue

  """
  for data_set_key,data_set_value in data_set.items():
    print("===============")
    print(data_set_key)             # Is one of "_l1_a", "_l2_a", "_l1_b", "_l2_b"
    #print(data_set_value)
    print("===============")
    # For each "_l1_a", "_l2_a", "_l1_b", "_l2_b"
    # get keys (for example: "ITU-Lambda1 = 18.0, ITU-Lambda2 = 21.0")
    # get values (number of measurements, mean crtt value, stddev of crtt values)
    for key,value in data_set_value.items():
      print(key)
      print(value)
  """
  l1_stddev = 3e-12 # 3 [pm]
  l2_stddev = 3e-12 # 3 [pm]
  fixed_lambda_stddev  = 3e-12 # 3 [pm]

  # name is the root file name. Get the directory where this file exists and 
  # open a "result.txt" file in this directory
  result_file = open((name + "_result.txt"),"w")
  result_file.write("# Proposed 3-Lambda Delta Delay measurements\n")
  result_file.write("# file: " + name + "\n")
  result_file.write("# fixed_lambda: " + str(fixed_lambda) + " = ITU-ch: "+str(fixed_itu_channel)+"\n")
  result_file.write("# Lambda 1, Lambda 2, CRTT(Lambda_1), CRTT(Lambda_2), Fixed Lambda, Alpha, Alpha SPEC format, Alpha StdDev, Alpha/Alpha_StdDev\n")

  for key,value in data_set["_l1_a"].items():
    # Get the ITU channel numbers used for lambda 1 and 2 and
    # convert them to wavelength in [m]
    key_lst=key.split(",")  
    ITU_l1 = float(key_lst[0][len("ITU Lambda1 = "):])
    ITU_l2 = float(key_lst[1][len("ITU Lambda2 = "):])
    l1 = itu_conv.itu_2_wavelength(ITU_l1)
    l2 = itu_conv.itu_2_wavelength(ITU_l2)
    
    # It can be seen (print statements below) that the difference in measured CRTT for
    # ("_l1_a" - "_l1_b") and ("_l2_a" - "_l2_b") is within the stddev of the measured
    # crtt values. This proves that the direction from which the tunable laser is set
    # doesn't  contribute to a systematic error.
    # Therefore the average over all measurements ("_l1_a" + "_l1_b")/2 is taken to be the 
    # value for lambda 1 (and the same for lambda 2) as well as the RMS value of the stddev.
    crtt_1_meas     =            data_set["_l1_a"][key][0]    + data_set["_l1_b"][key][0]
    crtt_2_meas     =            data_set["_l2_a"][key][0]    + data_set["_l2_b"][key][0]
    crtt_1a_1b_diff =           (data_set["_l1_a"][key][1]    - data_set["_l1_b"][key][1])
    crtt_2a_2b_diff =           (data_set["_l2_a"][key][1]    - data_set["_l2_b"][key][1])
    crtt_1          =           (data_set["_l1_a"][key][1]    + data_set["_l1_b"][key][1])/2
    crtt_2          =           (data_set["_l2_a"][key][1]    + data_set["_l2_b"][key][1])/2
    crtt_1_stddev   = numpy.sqrt(data_set["_l1_a"][key][2]**2 + data_set["_l1_b"][key][2]**2)
    crtt_2_stddev   = numpy.sqrt(data_set["_l2_a"][key][2]**2 + data_set["_l2_b"][key][2]**2)
    print("===============")
    print(key)
    print("===============")
    print("Lambda 1: ", l1)
    print("===============")
    print("Number of measurements: ", crtt_1_meas)
    print("CRTT a-b diff: ", crtt_1a_1b_diff)
    print("CRTT average: ", round(crtt_1))
    print("CRTT stddev: ", crtt_1_stddev)
    print("===============")
    print("Lambda 2: ", l2)
    print("===============")
    print("Number of measurements: ", crtt_2_meas)
    print("CRTT a-b diff: ", crtt_2a_2b_diff)
    print("CRTT average: ", round(crtt_2))
    print("CRTT stddev: ", crtt_2_stddev)

    # Calculate the (mean) value of alpha
    alpha = calc_alpha_l1(l1, l2, crtt_1, crtt_2, fixed_lambda)
    alphaspec = spec_alpha(alpha)

    # Calculate the stddev contribution to alpha of each of the parameters l1, l2, crtt_l1, crtt_l2 and fixed_lambda
    alpha_stddev_l1           = calc_alpha_l1(l1 + l1_stddev, l2, crtt_1, crtt_2, fixed_lambda) - alpha
    alpha_stddev_l2           = calc_alpha_l1(l1, l2 + l2_stddev, crtt_1, crtt_2, fixed_lambda) - alpha
    alpha_stddev_crtt_1       = calc_alpha_l1(l1, l2, crtt_1 + crtt_1_stddev, crtt_2, fixed_lambda) - alpha
    alpha_stddev_crtt_2       = calc_alpha_l1(l1, l2, crtt_1, crtt_2 + crtt_2_stddev, fixed_lambda) - alpha
    alpha_stddev_fixed_lambda = calc_alpha_l1(l1, l2, crtt_1, crtt_2, fixed_lambda + fixed_lambda_stddev) - alpha
    # Calculate the RMS stddev of alpha
    alpha_stddev = numpy.sqrt(alpha_stddev_l1**2 + alpha_stddev_l2**2 + alpha_stddev_crtt_1**2 + alpha_stddev_crtt_2**2 + alpha_stddev_fixed_lambda **2)
    print("===============")
    print("fixed_lambda: ", fixed_lambda)
    print("delta_l1: ", l1-fixed_lambda)
    print("delta_l2: ", l2-fixed_lambda)
    print("l1-l2: ", l1-l2)
    print("alpha                     ", alpha                    )
    print("alpha_stddev_l1           ", alpha_stddev_l1          )
    print("alpha_stddev_l2           ", alpha_stddev_l2          )
    print("alpha_stddev_crtt_1       ", alpha_stddev_crtt_1      )
    print("alpha_stddev_crtt_2       ", alpha_stddev_crtt_2      )
    print("alpha_stddev_fixed_lambda ", alpha_stddev_fixed_lambda)
    print("alpha_stddev              ", alpha_stddev             )
    
    print("===============")
    result_file.write(str(l1) + ", "  + str(l2) + ", "  + str(crtt_1) + ", "  + str(crtt_2) + ", "  + str(fixed_lambda) + ", "  + str(alpha) + ", "  + str(alphaspec) + ", "  + str(alpha_stddev) + ", " + str(alpha/alpha_stddev) + "\n")
  
sys.exit()
